/*
     This file is part of GNUnet.
     (C) 2003 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * Testbed CORE. This is the code that is plugged
 * into the GNUnet core to enable transport profiling.
 *
 * Todo:
 * - don't use system/wget to do HTTP registration
 * - support ssh in HTTP registration
 *
 * @author Ronaldo Alves Ferreira
 * @author Christian Grothoff
 * @author Murali Khrisna Ramanathan
 * @file applications/testbed/testbed.c
 **/


#include "testbed.h"

#include <sys/types.h>

#define DEBUG_TESTBED YES

/* */
static CoreAPIForApplication * coreAPI = NULL;

static void sendAcknowledgement(ClientHandle client,
				int ack) {
  if (OK != coreAPI->sendTCPResultToClient(client, ack)) {
    LOG(LOG_WARNING,
	"WARNING: could not send ack back to client.\n");
  }
}

/**
 * Handler that is called for "message not understood" cases.
 */
static void tb_undefined(ClientHandle client,
			 TESTBED_CS_MESSAGE * msg) {
  LOG(LOG_WARNING,
      "WARNING: received unknown testbed message of type %u\n",
      ntohl(msg->msgType));
}

/**
 * Connect to another peer.
 */
static void tb_ADD_PEER(ClientHandle client, 
			TESTBED_CS_MESSAGE * msg) {
  p2p_HEADER noise;
  TESTBED_ADD_PEER_MESSAGE * hm 
    = (TESTBED_ADD_PEER_MESSAGE*) msg;
  
  LOG(LOG_DEBUG, 
      "DEBUG: tb_ADD_PEER\n");
  if (sizeof(TESTBED_ADD_PEER_MESSAGE) >
      ntohs(msg->header.size) ) {
    LOG(LOG_ERROR,
	"ERROR: size of ADD_PEER message is too short.  Ignoring.\n");
    return;
  }
  if (HELO_Message_size(&hm->helo) !=
      ntohs(msg->header.size) - sizeof(TESTBED_CS_MESSAGE) ) {
    LOG(LOG_ERROR,
	"ERROR: size of ADD_PEER message is wrong.  Ignoring.\n");
    return;
  }
  
  coreAPI->bindAddress(&hm->helo);
  noise.size = htons(sizeof(p2p_HEADER));
  noise.requestType = htons(p2p_PROTO_NOISE);
  coreAPI->sendToNode(&hm->helo.senderIdentity,
		      &noise,
		      EXTREME_PRIORITY, 
		      0);
  sendAcknowledgement(client, OK);
}

/**
 * Disconnect from another peer.
 */
static void tb_DEL_PEER(ClientHandle client,
			TESTBED_DEL_PEER_MESSAGE * msg) {
  coreAPI->disconnectFromPeer(&msg->host);
  sendAcknowledgement(client, OK);
}

/**
 * Disconnect from all other peers.
 */
static void tb_DEL_ALL_PEERS(ClientHandle client,
			     TESTBED_DEL_ALL_PEERS_MESSAGE * msg) {
  coreAPI->disconnectPeers(); 
  sendAcknowledgement(client, OK);
}

/**
 * Get a HELO message for this peer.
 */
static void tb_GET_HELO(ClientHandle client,
			TESTBED_GET_HELO_MESSAGE * msg) {
  HELO_Message * helo;
  unsigned int proto = ntohs(msg->proto);
  
  if (SYSERR == coreAPI->identity2Helo(coreAPI->myIdentity, 
				       proto,
				       NO, 
				       &helo)) {
    LOG(LOG_WARNING, 
	"WARNING: TESTBED could not generate HELO"	\
	"message for protocol %u\n",
	proto);
    sendAcknowledgement(client, SYSERR);
  } else {
    TESTBED_HELO_MESSAGE * reply 
      = MALLOC(ntohs(helo->header.size) + sizeof(TESTBED_CS_MESSAGE));
    reply->header.header.size 
      = htons(ntohs(helo->header.size) + sizeof(TESTBED_CS_MESSAGE));
    reply->header.header.tcpType 
      = htons(TESTBED_CS_PROTO_REPLY);
    reply->header.msgType
      = htonl(TESTBED_HELO_RESPONSE);
    memcpy(&reply->helo, 
	   helo, 
	   ntohs(helo->header.size));
    coreAPI->sendToClient(client,
			  &reply->header.header);
    LOG(LOG_DEBUG,
	"DEBUG: tb_GET_HELO: returning from writeToSocket\n");
    FREE(helo);
    FREE(reply);
  }    
}

/**
 * Set a trust value.
 */
static void tb_SET_TVALUE(ClientHandle client,
			  TESTBED_SET_TVALUE_MESSAGE * msg) {
  int trust, chg;
  
  trust = ntohl(msg->trust);
  chg = coreAPI->changeTrust(&msg->otherPeer, trust);
  if (chg != trust) {
    LOG(LOG_WARNING,
	"WARNING: trust change=%d, required=%d\n",
	chg,
	trust);
  }
  sendAcknowledgement(client, OK);
}	
		  
/**
 * Get a trust value.
 */
static void tb_GET_TVALUE(ClientHandle client,
			  TESTBED_GET_TVALUE_MESSAGE * msg) {    
  unsigned int trust;
  
  trust = coreAPI->getTrust(&msg->otherPeer);
  sendAcknowledgement(client, trust);
}	

/**
 * Change the bandwidth limitations.
 */
static void tb_SET_BW(ClientHandle client,
		      TESTBED_SET_BW_MESSAGE * msg) {
  LOG(LOG_DEBUG,
      "DEBUG: gnunet-testbed: tb_SET_BW\n");
  setConfigurationInt("LOAD",
		      "MAXNETDOWNBPSTOTAL", 
		      ntohl(msg->in_bw));
  setConfigurationInt("LOAD",
		      "MAXNETUPBPSTOTAL", 
		      ntohl(msg->out_bw));
  triggerGlobalConfigurationRefresh();
  sendAcknowledgement(client, OK);
}		  

/**
 * Load an application module.
 */
static void tb_LOAD_MODULE(ClientHandle client,
			   TESTBED_CS_MESSAGE * msg) {
  unsigned short size;
  char * name;
  int ok;
  
  size = ntohs(msg->header.size);
  if (size <= sizeof(TESTBED_CS_MESSAGE) ) {
    LOG(LOG_WARNING,
	"WARNING: received invalid LOAD_MODULE message\n");
    return;
  } 
  
  if (! testConfigurationString("TESTBED",
				"ALLOW_MODULE_LOADING",
				"YES")) {
    sendAcknowledgement(client, SYSERR);
    return;
  }
  
  name = STRNDUP(&((TESTBED_LOAD_MODULE_MESSAGE*)msg)->modulename[0],
		 size - sizeof(TESTBED_CS_MESSAGE));
  if (strlen(name) == 0) {
    LOG(LOG_WARNING,
	"WARNING: received invalid LOAD_MODULE"	\
	"message (empty module name)\n");
    return;
  }
  ok = coreAPI->loadApplicationModule(name);
  if (ok != OK)
    LOG(LOG_WARNING,
	"WARNING: loading module failed.  Notifying client.\n");
  FREE(name);
  sendAcknowledgement(client, ok);
}

/**
 * Unload an application module.
 */
static void tb_UNLOAD_MODULE(ClientHandle client,
			     TESTBED_CS_MESSAGE * msg) {
  unsigned short size;
  char * name;
  int ok;
  
  size = ntohs(msg->header.size);
  if (size <= sizeof(TESTBED_CS_MESSAGE) ) {
    LOG(LOG_WARNING,
	"WARNING: received invalid UNLOAD_MODULE message\n");
    return;
  }
  
  if (! testConfigurationString("TESTBED",
				"ALLOW_MODULE_LOADING",
				"YES")) {      
    sendAcknowledgement(client, SYSERR);
    return;
  }
  
  name = STRNDUP(&((TESTBED_UNLOAD_MODULE_MESSAGE*)msg)->modulename[0],
		 size - sizeof(TESTBED_CS_MESSAGE));
  if (strlen(name) == 0) {
    LOG(LOG_WARNING,
	"WARNING: received invalid UNLOAD_MODULE"	\
	"message (empty module name)\n");
    return;
  }
  ok = coreAPI->unloadApplicationModule(name);
  if (ok != OK)
    LOG(LOG_WARNING,
	"WARNING: unloading module failed.  Notifying client.\n");
  FREE(name);
  sendAcknowledgement(client, ok);
}

/**
 * Set the reliability of the inbound and outbound transfers for this
 * peer (by making it drop a certain percentage of the messages at
 * random).
 */
static void tb_SET_LOSS_RATE(ClientHandle client,
			     TESTBED_SET_LOSS_RATE_MESSAGE * msg) {
  coreAPI->setPercentRandomInboundDrop
    (ntohl(msg->percentageLossInbound));
  coreAPI->setPercentRandomOutboundDrop
    (ntohl(msg->percentageLossOutbound));
  sendAcknowledgement(client, OK);
}

/**
 * Set the reliability of the inbound and outbound transfers for this
 * peer (by making it drop a certain percentage of the messages at
 * random).
 */
static void tb_DISABLE_AUTOCONNECT(ClientHandle client,
				   TESTBED_DISABLE_AUTOCONNECT_MESSAGE * msg) {
  FREENONNULL(setConfigurationString("GNUNETD",
				     "DISABLE-AUTOCONNECT",
				     "YES"));
  triggerGlobalConfigurationRefresh();
  sendAcknowledgement(client, OK);
}

/**
 * Set the reliability of the inbound and outbound transfers for this
 * peer (by making it drop a certain percentage of the messages at
 * random).
 */
static void tb_ENABLE_AUTOCONNECT(ClientHandle client,
				  TESTBED_ENABLE_AUTOCONNECT_MESSAGE * msg) {
  FREENONNULL(setConfigurationString("GNUNETD",
				     "DISABLE-AUTOCONNECT",
				     "NO"));
  triggerGlobalConfigurationRefresh();
  sendAcknowledgement(client, OK);
}

/**
 * Set the reliability of the inbound and outbound transfers for this
 * peer (by making it drop a certain percentage of the messages at
 * random).
 */
static void tb_DISABLE_HELO(ClientHandle client,
			    TESTBED_DISABLE_HELO_MESSAGE * msg) {
  FREENONNULL(setConfigurationString("NETWORK",
				     "DISABLE-ADVERTISEMENTS",
				     "YES"));
  FREENONNULL(setConfigurationString("NETWORK",
				     "HELOEXCHANGE",
				     "NO"));
  triggerGlobalConfigurationRefresh();
  sendAcknowledgement(client, OK);
}

/**
 * Set the reliability of the inbound and outbound transfers for this
 * peer (by making it drop a certain percentage of the messages at
 * random).
 */
static void tb_ENABLE_HELO(ClientHandle client,
			   TESTBED_ENABLE_HELO_MESSAGE * msg) {
  FREENONNULL(setConfigurationString("NETWORK",
				     "DISABLE-ADVERTISEMENTS",
				     "NO"));
  FREENONNULL(setConfigurationString("NETWORK",
				     "HELOEXCHANGE",
				     "YES"));
  triggerGlobalConfigurationRefresh();
  sendAcknowledgement(client, OK);
}

/**
 * Allow only certain peers to connect.
 */
static void tb_ALLOW_CONNECT(ClientHandle client,
			     TESTBED_ALLOW_CONNECT_MESSAGE * msg) {
  char * value;
  unsigned short size;
  unsigned int count;
  unsigned int i;
  HexName hex;

  size = ntohs(msg->header.header.size);
  if (size <= sizeof(TESTBED_CS_MESSAGE) ) {
    LOG(LOG_WARNING,
	"WARNING: received invalid ALLOW_CONNECT message\n");
    return;
  }
  count = (size - sizeof(TESTBED_CS_MESSAGE)) / sizeof(HostIdentity);
  if (count * sizeof(HostIdentity) + sizeof(TESTBED_CS_MESSAGE) != size) {
    LOG(LOG_WARNING,
	"WARNING: received invalid ALLOW_CONNECT message\n");
    return;
  }
  if (count == 0) {
    value = NULL;
  } else {
    value = MALLOC(count * sizeof(HostIdentity)*2 + 1);
    value[0] = '\0';
    for (i=0;i<count;i++) {
      hash2hex(&msg->peers[i].hashPubKey,
	       &hex);
      strcat(value, (char*)&hex);
    }
  }
  FREENONNULL(setConfigurationString("GNUNETD",
				     "LIMIT-ALLOW",
				     value));
  FREENONNULL(value);
  triggerGlobalConfigurationRefresh();
  sendAcknowledgement(client, OK);
}

/**
 * Deny certain peers the right to connect.
 */
static void tb_DENY_CONNECT(ClientHandle client,
			    TESTBED_DENY_CONNECT_MESSAGE * msg) {
  char * value;
  unsigned short size;
  unsigned int count;
  unsigned int i;
  HexName hex;

  size = ntohs(msg->header.header.size);
  if (size <= sizeof(TESTBED_CS_MESSAGE) ) {
    LOG(LOG_WARNING,
	"WARNING: received invalid DENY_CONNECT message\n");
    return;
  }
  count = (size - sizeof(TESTBED_CS_MESSAGE)) / sizeof(HostIdentity);
  if (count * sizeof(HostIdentity) + sizeof(TESTBED_CS_MESSAGE) != size) {
    LOG(LOG_WARNING,
	"WARNING: received invalid DENY_CONNECT message\n");
    return;
  }
  if (count == 0) {
    value = NULL;
  } else {
    value = MALLOC(count * sizeof(HostIdentity)*2 + 1);
    value[0] = '\0';
    for (i=0;i<count;i++) {
      hash2hex(&msg->peers[i].hashPubKey,
	       &hex);
      strcat(value, (char*)&hex);
    }
  }
  FREENONNULL(setConfigurationString("GNUNETD",
				     "LIMIT-DENY",
				     value));
  FREENONNULL(value);
  triggerGlobalConfigurationRefresh();
  sendAcknowledgement(client, OK);
}

/**
 * Information about processes that we have forked off.
 */
typedef struct {
  /** the unique identifier of the PI */
  unsigned int uid;
  /** the PID of the process */
  pid_t pid; 
  /** stdout and stderr of the process */
  int outputPipe;
  /** thread that reads the output of the process */
  pthread_t reader;
  /** how many bytes of output did the process produce? */
  int outputSize;
  /** the output of the process */ 
  char * output; 
  /** did the process exit? (YES/NO) */
  int hasExited; 
  /** if the process did exit, what was the status? (or errno of execve) */
  int exitStatus;  
} ProcessInfo;

static unsigned int uidCounter = 0;

/**
 * The process table.
 */
static ProcessInfo ** pt = NULL;

/**
 * Number of entries in the process table.
 */
static int ptSize = 0;

/**
 * Lock for accessing the PT
 */
static Mutex lock;

/**
 * Thread that captures the output from a child-process
 * and stores it in the process info.
 */
static int pipeReaderThread(ProcessInfo * pi) {
  int ret = 1;
  char * buffer;
#define PRT_BUFSIZE 65536
  buffer = MALLOC(PRT_BUFSIZE);
  while (ret > 0) {
    ret = read(pi->outputPipe,
	       buffer,
	       PRT_BUFSIZE);
    MUTEX_LOCK(&lock);
    if (pi->outputSize == -1)
      break;
    GROW(pi->output,
	 pi->outputSize,
	 pi->outputSize + ret);
    memcpy(&pi->output[pi->outputSize-ret],
	   buffer,
	   ret);
    MUTEX_UNLOCK(&lock);
  }
  CLOSE(pi->outputPipe);
  MUTEX_LOCK(&lock);
  if (pi->outputSize == -1) {
    int i;
    /* the main testbed handler has already given
       up on this pi; we're responsible to free
       the associated resources! */
    if (0 != pthread_detach(pi->reader))
      LOG(LOG_WARNING,
	  "WARNING: could not detach myself: %s\n",
	  strerror(errno));
    FREENONNULL(pi->output);
    for (i=0;i<ptSize;i++)
      if (pt[i] == pi)
	break;
    if (i == ptSize) {
      LOG(LOG_WARNING,
	  "WARNING: could not find pi entry in PT!\n");
    } else {
      pt[i] = pt[ptSize-1];
      GROW(pt,
	   ptSize,
	   ptSize-1);
    }
    FREE(pi);
  } else {
    if (pi->pid != waitpid(pi->pid,
			   &pi->exitStatus,
			   0)) {
      LOG(LOG_WARNING,
	  "WARNING: waidpid failed: %s\n",
	  strerror(errno));
      pi->exitStatus = errno;
    }
    pi->hasExited = YES;
  }  
  MUTEX_UNLOCK(&lock);
  return 0;
}

/**
 * Execute a command.
 */
static void tb_EXEC(ClientHandle client,
		    TESTBED_CS_MESSAGE * msg) {
  int argc;
  int argc2;
  int fd[2];
  char ** argv;
  unsigned short size;
  unsigned int uid;
  int pos;
  TESTBED_EXEC_MESSAGE * emsg;
  ProcessInfo * pi;

  emsg = (TESTBED_EXEC_MESSAGE*)msg;
  size = htons(msg->header.size);
  if ( (size <= sizeof(TESTBED_CS_MESSAGE)) ||
       (emsg->commandLine[size-sizeof(TESTBED_CS_MESSAGE)] != '\0') ) {
    LOG(LOG_WARNING,
	"WARNING: received invalid EXEC message.\n");
    sendAcknowledgement(client, SYSERR);
    return;
  }
  size -= sizeof(TESTBED_CS_MESSAGE);
  argc = 0;
  for (pos=0;pos<size;pos++)
    if (emsg->commandLine[pos] == '\0')
      argc++;
  argc2 = argc;
  if (0 != pipe(fd)) {
    LOG(LOG_WARNING,
	"WARNING: could not create pipe: %s.\n",
	strerror(errno));
    sendAcknowledgement(client, SYSERR);
    return;
  }
  argv = MALLOC(sizeof(char*)*argc);
  for (pos=size-2;pos>=0;pos--)
    if (emsg->commandLine[pos] == '\0')
      argv[--argc2] = STRDUP(&emsg->commandLine[pos+1]);
  argv[--argc2] = STRDUP(&emsg->commandLine[0]);
  MUTEX_LOCK(&lock);
  pi = MALLOC(sizeof(ProcessInfo));
  pi->pid = fork();
  if (pi->pid == 0) {
    /* make pipe stdout/stderr */
    CLOSE(fd[0]);
    CLOSE(0);
    CLOSE(1);
    CLOSE(2);
    if (-1 == dup2(fd[0], 1))
      LOG(LOG_ERROR,
	  "ERROR: could not dup2 pipe to be stdout (%s)!\n",
	  strerror(errno));
    if (-1 == dup2(fd[0], 2))
      LOG(LOG_ERROR,
	  "ERROR: could not dup2 pipe to be stdout (%s)!\n",
	  strerror(errno));
    execvp(argv[0],
	   &argv[1]);
    exit(errno);
  }
  CLOSE(fd[1]);
  for (pos=0;pos<argc;pos++)
    FREE(argv[pos]);
  FREE(argv);
  if (pi->pid == -1) {
    LOG(LOG_WARNING,
	"WARNING: fork failed: %s\n",
	strerror(errno));
    FREE(pi);
    CLOSE(fd[0]);
    sendAcknowledgement(client, SYSERR);
    return;    
  }
  pi->uid = uidCounter++;
  pi->outputPipe = fd[0];
  pi->outputSize = 0;
  pi->output = NULL;
  pi->hasExited = NO;
  pi->exitStatus = 0;
  PTHREAD_CREATE(&pi->reader,
		 (PThreadMain) &pipeReaderThread,
		 pi,
		 4*1024);
  GROW(pt,
       ptSize,
       ptSize+1);
  pt[ptSize-1] = pi;  
  uid = pi->uid;
  MUTEX_UNLOCK(&lock);
  sendAcknowledgement(client, uid);
}

/**
 * Send a signal to a process.
 */
static void tb_SIGNAL(ClientHandle client,
		      TESTBED_SIGNAL_MESSAGE * msg) {
  int ret;
  int i;
  unsigned int uid;
  int sig;

  ret = SYSERR;
  uid = ntohl(msg->pid);
  sig = ntohl(msg->signal);
  MUTEX_LOCK(&lock);
  for (i=0;i<ptSize;i++) {
    ProcessInfo * pi;

    pi = pt[i];
    if (pi->uid == uid) {
      if (sig == 0) {	
	ret = OK;
	pi->outputSize = -1;
	FREENONNULL(pi->output);
	pi->output = NULL;
	if (pi->hasExited != NO) {
	  void *unused;
	  if (0 != pthread_join(pi->reader,
				&unused))
	    LOG(LOG_WARNING,
		"WARNING: pthread_join failed: %s\n",
		strerror(errno));
	  FREE(pi);
	  pt[i] = pt[ptSize-1];
	  GROW(pt,
	       ptSize,
	       ptSize-1);
	  break;
	}
      } else {
	if (pi->hasExited == NO) {      
	  if (0 == kill(pi->pid, ntohl(msg->signal)))
	    ret = OK;
	  else
	    LOG(LOG_WARNING,
		"WARNING: could not send signal to process %d: %s\n",
		pi->pid,
		strerror(errno));
	}
	break;
      }
    }
  }
  MUTEX_UNLOCK(&lock);
  sendAcknowledgement(client, ret);
}

/**
 * Get the output of a process.
 */
static void tb_GET_OUTPUT(ClientHandle client,
			  TESTBED_GET_OUTPUT_MESSAGE * msg) {
  int i;
  unsigned int uid;

  uid = ntohl(msg->pid);
  MUTEX_LOCK(&lock);
  for (i=0;i<ptSize;i++) {
    ProcessInfo * pi;

    pi = pt[i];
    if (pi->uid == uid) {
      int pos;
      TESTBED_OUTPUT_REPLY_MESSAGE * msg;

      msg = MALLOC(65532);
      msg->header.header.tcpType 
	= htons(TESTBED_CS_PROTO_REPLY);
      msg->header.msgType 
	= htonl(TESTBED_OUTPUT_RESPONSE);
      
      sendAcknowledgement(client, pi->outputSize);
      pos = 0;
      while (pos < pi->outputSize) {	
	int run = pi->outputSize - pos;
	if (run > 65532 - sizeof(TESTBED_OUTPUT_REPLY_MESSAGE))
	  run = 65532 - sizeof(TESTBED_OUTPUT_REPLY_MESSAGE);
	msg->header.header.size
	  = htons(run+sizeof(TESTBED_OUTPUT_REPLY_MESSAGE));
	memcpy(&msg->data[0],
	       &pi->output[pos],
	       run);
	coreAPI->sendToClient(client,
			      &msg->header.header);
	pos += run;
      }
      FREE(msg);
      MUTEX_UNLOCK(&lock);
      return;
    }
  } 
  MUTEX_UNLOCK(&lock);
  sendAcknowledgement(client, SYSERR);
}

/**
 * The client is uploading a file to this peer.
 */
static void tb_UPLOAD_FILE(ClientHandle client,
			   TESTBED_UPLOAD_FILE_MESSAGE * msg) {
  int ack, size;
  char * filename, *gnHome, *s;
  char * end;
  FILE *outfile;
  
  LOG(LOG_DEBUG, 
      "DEBUG: tb_UPLOAD_FILE\n");
  if (sizeof(TESTBED_UPLOAD_FILE_MESSAGE) > ntohs(msg->header.header.size)) {
    LOG(LOG_ERROR,
	"ERROR: size of UPLOAD_FILE message is too short. Ignoring.\n");
    sendAcknowledgement(client, SYSERR);
    return;
  }
  end = &((char*)msg)[ntohs(msg->header.header.size)];
  s = filename = msg->buf;
  while ( (*s) && (s != end) ) {
    if (*s == '.' && *(s+1) == '.') {
      LOG(LOG_ERROR,
	  "ERROR: \'..\' is not allowed in file name (%s).\n",
	  filename);
      return;
    }
    s++;
  }
  if (s == filename) {
    /* filename empty, not allowed! */
    LOG(LOG_ERROR,
	"ERROR: empty filename for UPLOAD_FILE message is illegal!\n"); 
    sendAcknowledgement(client, SYSERR);
    return;
  }
  if (s == end) {
    /* filename empty, not allowed! */
    LOG(LOG_ERROR,
	"ERROR: filename for UPLOAD_FILE message is not null-terminated (illegal!)\n"); 
    sendAcknowledgement(client, SYSERR);
    return;
  }
  gnHome = getFileName("",
		       "GNUNETD_HOME",
		       "Configuration file must specify a "
		       "directory for GNUnet to store "
		       "per-peer data under %s%s\n");
  
  filename = MALLOC(strlen(filename) + strlen(gnHome) + 2); /*2: /, \0 */ 
  strcpy(filename, gnHome);
  strcat(filename, "/");
  strcat(filename, msg->buf);
  if (htonl(msg->type) == TESTBED_FILE_DELETE) {
    if (remove(filename) && errno != ENOENT) {
      LOG(LOG_WARNING,
	  "WARNING: could not remove file %s (%s)\n",
	  filename,
	  STRERROR(errno));
      ack = SYSERR;
    } else
      ack = OK;
    FREE(filename);
    sendAcknowledgement(client, ack);
    return;
  }   
  if (htonl(msg->type) != TESTBED_FILE_APPEND) {
    LOG(LOG_ERROR,
	"ERROR: Invalid message received at UPLOAD_FILE.\n");
    FREE(filename);
    return;
  }   
  outfile = FOPEN(filename, "ab");
  if (outfile == NULL) {
    /* Send nack back to control point. */
    LOG(LOG_ERROR, 
	"ERROR: could not create file %s\n",
	filename);
    sendAcknowledgement(client, SYSERR);
    FREE(filename);
    return;
  }
  FREE(filename);
  s    = msg->buf + strlen(msg->buf) + 1; /* \0 added */
  size = ntohs(msg->header.header.size) -
    sizeof(TESTBED_UPLOAD_FILE_MESSAGE) - (strlen(msg->buf)+1);
  if (GN_FWRITE(s, 1, size, outfile) != size)
    ack = SYSERR;
  else
    ack = OK;
  fclose(outfile);
  sendAcknowledgement(client, ack);
}

/**
 * General type of a message handler.
 */			  
typedef void (*THandler)(ClientHandle client,
			 TESTBED_CS_MESSAGE * msg);

/**
 * @brief Entry in the handlers array that describes a testbed message handler.
 */
typedef struct HD_ {
    /**
     * function that handles these types of messages 
     **/
    THandler handler; 

    /**
     * Expected size of messages for this handler.  Checked by caller.  Use
     * 0 for variable size, in that case, the handler must check.
     */
    unsigned short expectedSize; 

    /**
     * Textual description of the handler for debugging 
     */
    char * description; 
    
    /**
     * The message-ID of the handler.  Used only for checking that 
     * the handler array matches the message IDs defined in testbed.h.
     * Must be equal to the index in the handler array that yields
     * this entry.
     */
    unsigned int msgId; 
} HD;

/* some macros to make initializing the handlers array extremely brief. */

#define TBDENTRY(a)  {(THandler)&tb_##a, 0, "##a##", TESTBED_##a }
#define TBSENTRY(a)  {(THandler)&tb_##a, sizeof(TESTBED_##a##_MESSAGE),\
		      "##a##", TESTBED_##a}

/**
 * The array of message handlers.  Add new handlers here.
 */
static HD handlers[] = {
  TBSENTRY(undefined),	/* For IDs that should never be received */
  TBDENTRY(ADD_PEER),	/* RF: Why was this as TBDENTRY? Because HELO is variable size! */
  TBSENTRY(DEL_PEER), 
  TBSENTRY(DEL_ALL_PEERS),
  TBSENTRY(GET_HELO), 
  TBSENTRY(SET_TVALUE), 
  TBSENTRY(GET_TVALUE), 
  TBSENTRY(undefined), 
  TBSENTRY(SET_BW), 
  TBSENTRY(SET_LOSS_RATE),
  TBDENTRY(LOAD_MODULE),
  TBDENTRY(UNLOAD_MODULE),
  TBDENTRY(UPLOAD_FILE),
  TBSENTRY(DISABLE_HELO),
  TBSENTRY(ENABLE_HELO),
  TBSENTRY(DISABLE_AUTOCONNECT),
  TBSENTRY(ENABLE_AUTOCONNECT),
  TBDENTRY(ALLOW_CONNECT),
  TBDENTRY(DENY_CONNECT),
  TBDENTRY(EXEC),
  TBSENTRY(SIGNAL),
  TBSENTRY(GET_OUTPUT),
  { NULL, 0, NULL, 0 },	/* this entry is used to ensure that
			   a wrong TESTBED_MAX_MSG will abort
			   insted of possibly segfaulting.
			   This must always be the LAST entry. */
};

/**
 * Number of entries in handlers array.  Checked automatically when the
 * module is initialized.
 */
#define TESTBED_MAX_MSG 23

/**
 * Global handler called by the GNUnet core.  Does the demultiplexing
 * on the testbed-message type.
 */
static void csHandleTestbedRequest(ClientHandle client,
				   CS_HEADER * message) {
  TESTBED_CS_MESSAGE * msg;
  unsigned short size;
  unsigned int id;
  
#if DEBUG_TESTBED
  LOG(LOG_DEBUG, 
      "DEBUG: TESTBED handleTestbedRequest\n");
#endif
  size = ntohs(message->size);
  if (size < sizeof(TESTBED_CS_MESSAGE)) {
    LOG(LOG_WARNING,
	"WARNING: received invalid testbed message of size %u\n",
	size);
    return;
  }
  msg = (TESTBED_CS_MESSAGE *)message;
  id = ntohl(msg->msgType);
  if (id < TESTBED_MAX_MSG) {
    if ( (handlers[id].expectedSize == 0) ||
	 (handlers[id].expectedSize == size) ) {
      LOG(LOG_DEBUG, 
	  "DEBUG: TESTBED received %s message:(%u)\n",
	  handlers[id].description, id);
      
      handlers[id].handler(client, msg);
      
    } else {
      LOG(LOG_ERROR,
	  "ERROR: received message of type %u but"	\
	  "unexpected size %u, expected %u\n",
	  id, 
	  size,
	  handlers[id].expectedSize);
    }
  } else {
    tb_undefined(client, msg);
  }
} 

/**
 * Register this testbed peer with the central testbed server.
 * Yes, the testbed has a central server.  There's nothing wrong
 * with that.  It's a testbed.
 */
static void httpRegister(char * cmd) {
  char * reg;
  reg = getConfigurationString("TESTBED",
			       "REGISTERURL");
  if (reg != NULL) {
    char * wget;
    char * trusted;
    int i;
    
    trusted = getConfigurationString("NETWORK",
				     "TRUSTED");
    if (trusted == NULL)
      trusted = STRDUP("127.0.0.0/8;");
    i = 0;
    while (trusted[i] != '\0') {
      if (trusted[i] == ';')
	trusted[i] = '@';
      i++;
    }
	 
    wget = MALLOC(128 + strlen(reg) + strlen(trusted));
    sprintf(wget,
	    "wget -O /dev/null %s/%s.php3?trusted=%s &> /dev/null",
	    reg,
	    cmd,
	    trusted);
    FREE(trusted);
    FREE(reg);
    LOG(LOG_DEBUG,
	"DEBUG: testbed registration: %s\n",
	wget);
    if (0 != system(wget)) { /* FIXME: this is lazy, factor http code from heloexchange
				to also use the proxy settings and don't rely on wget. */
      LOG(LOG_ERROR,
	  "ERROR: command %s failed, testbed not registered for service.\n",
	  wget);
    }
    FREE(wget);
  } else {
    LOG(LOG_DEBUG,
	"DEBUG: no testbed URL given, not registered.\n");
  }
} 

/**
 * Initialize the TESTBED module. This method name must match
 * the library name (libgnunet_XXX => initialize_XXX).
 * @return SYSERR on errors
 **/
int initialize_testbed_protocol(CoreAPIForApplication * capi) {
  int i;
  int ok = OK;
  
  /* some checks */
  for (i=0;i<TESTBED_MAX_MSG;i++)
    if ( (handlers[i].msgId != i) &&
	 (handlers[i].handler != &tb_undefined) )
      errexit("FATAL: Assertion failed: "
	      "Malformed handlers array in testbed.c. "
	      "Aborting. (%d)\n", i);
  if (handlers[TESTBED_MAX_MSG].handler != NULL)
    errexit("FATAL: Assertion failed: "
	    "TESTBED_MAX_MSG in testbed.c is wrong."
	    "Aborting.\n");
  MUTEX_CREATE(&lock);
  LOG(LOG_DEBUG,
      "DEBUG: TESTBED registering handler %d!\n",
      TESTBED_CS_PROTO_REQUEST);
  coreAPI = capi;
  if (SYSERR == capi->registerClientHandler(TESTBED_CS_PROTO_REQUEST,
					    (CSHandler)&csHandleTestbedRequest))
    ok = SYSERR;
  httpRegister("startup");
  return ok;
}

/**
 * Shutdown the testbed module.
 */
void done_testbed_protocol() {
  int i;

  /* kill all child-processes */
  for (i=0;i<ptSize;i++) {
    ProcessInfo * pi;
    void * unused;

    pi = pt[i];
    if (pi->hasExited != NO)
      kill(pi->pid, SIGKILL);    
    if (0 != pthread_join(pi->reader,
			  &unused))
      LOG(LOG_WARNING,
	  "WARNING: pthread_join failed: %s\n",
	  strerror(errno));
    FREENONNULL(pi->output);
    FREE(pi);    
  }
  GROW(pt,
       ptSize,
       0);
  
  httpRegister("shutdown");
  MUTEX_DESTROY(&lock);
  LOG(LOG_DEBUG,
      "DEBUG: TESTBED unregistering handler %d\n",
      TESTBED_CS_PROTO_REQUEST);
  coreAPI->unregisterClientHandler(TESTBED_CS_PROTO_REQUEST,
				   (CSHandler)&csHandleTestbedRequest);
  coreAPI = NULL;
}

/* end of testbed.c */
